/* $Id: cpl_imagelist_io_body.h,v 1.17 2006/11/15 15:04:19 yjung Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Type dependent macros */
#if CPL_CLASS == CPL_CLASS_DOUBLE
#define CPL_PTYPE PTYPE_DOUBLE
#define CPL_IMAGE_GET cpl_image_get_data_double
#define CPL_TYPE_T CPL_TYPE_DOUBLE
#define CPL_TYPE double
#define CPL_IMAGE_WRAP cpl_image_wrap_double
#define CPL_QFITS_BUF dbuf

#elif CPL_CLASS == CPL_CLASS_FLOAT
#define CPL_PTYPE PTYPE_FLOAT
#define CPL_IMAGE_GET cpl_image_get_data_float
#define CPL_TYPE_T CPL_TYPE_FLOAT
#define CPL_TYPE float
#define CPL_IMAGE_WRAP cpl_image_wrap_float
#define CPL_QFITS_BUF fbuf

#elif CPL_CLASS == CPL_CLASS_INT
#define CPL_PTYPE PTYPE_INT
#define CPL_IMAGE_GET cpl_image_get_data_int
#define CPL_TYPE_T CPL_TYPE_INT
#define CPL_TYPE int
#define CPL_IMAGE_WRAP cpl_image_wrap_int
#define CPL_QFITS_BUF ibuf

#elif CPL_CLASS == CPL_CLASS_BIN
#define CPL_IMAGE_GET cpl_mask_get_data

#else
#undef CPL_PTYPE
#undef CPL_IMAGE_GET
#undef CPL_TYPE_T
#undef CPL_TYPE
#undef CPL_QFITS_BUF
#endif


#if CPL_OPERATION == CPL_IMLIST_IO_LOAD

    cpl_imagelist   *   imlist;
    cpl_image       *   image = NULL;
    qfitsloader         ql ;
    int                 nplanes ;
    int                 next;
    int                 i ;

    /* Test entries */
    cpl_ensure(filename != NULL,   CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(extnum   >= 0,      CPL_ERROR_ILLEGAL_INPUT, NULL);
    next = qfits_query_n_ext((char*)filename);
    cpl_ensure(next     >= 0,      CPL_ERROR_FILE_IO, NULL);
    cpl_ensure(next     >= extnum, CPL_ERROR_ACCESS_OUT_OF_RANGE, NULL);

    /* Initialize a FITS loader */
    ql.filename = (char *) filename ;
    ql.xtnum    = extnum ;
    ql.pnum     = 0 ;
    ql.map      = 0 ;
    ql.ptype    = CPL_PTYPE ;
    cpl_ensure(qfitsloader_init(&ql)==0,CPL_ERROR_BAD_FILE_FORMAT, NULL);
    nplanes = ql.np ;

    cpl_ensure(nplanes > 0, CPL_ERROR_DATA_NOT_FOUND, NULL);

    /* Create the image list */
    imlist = cpl_imagelist_new() ;

    /* Loop on all planes */
    for (i=0 ; i < nplanes ; i++) {
        ql.pnum     = i ;
        cpl_ensure(qfits_loadpix(&ql) == 0, CPL_ERROR_BAD_FILE_FORMAT, NULL);
        image = CPL_IMAGE_WRAP(ql.lx, ql.ly, ql.CPL_QFITS_BUF) ;
        if (image == NULL) break;
        
        if (cpl_imagelist_set(imlist, image, i)) break;
        image = NULL;
    }
    if (imlist->ni < nplanes) {
        cpl_image_delete(image);
        cpl_imagelist_delete(imlist);
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }
    
    /* All images in the created imagelist have the same size and type */
    assert( cpl_imagelist_is_uniform(imlist) == 0);
    
    /* Return */
    return imlist ;

#elif CPL_OPERATION == CPL_IMLIST_IO_LOAD_WIN

    cpl_imagelist   *   imlist;
    cpl_image       *   image = NULL;
    qfitsloader         ql ;
    int                 nplanes ;
    int                 next;
    int                 i ;

    /* Test entries */
    cpl_ensure(filename,         CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(extnum >= 0,      CPL_ERROR_ILLEGAL_INPUT, NULL);
    next = qfits_query_n_ext((char*)filename);
    cpl_ensure(next   >= 0,      CPL_ERROR_FILE_IO, NULL);
    cpl_ensure(next   >= extnum, CPL_ERROR_ACCESS_OUT_OF_RANGE, NULL);

    /* Initialize a FITS loader */
    ql.filename = (char *) filename ;
    ql.xtnum    = extnum ;
    ql.pnum     = 0 ;
    ql.map      = 0 ;
    ql.ptype    = CPL_PTYPE ;
    cpl_ensure(qfitsloader_init(&ql)==0,CPL_ERROR_BAD_FILE_FORMAT, NULL);
    nplanes = ql.np ;

    cpl_ensure(nplanes > 0, CPL_ERROR_DATA_NOT_FOUND, NULL);

    /* Create the image list */
    imlist = cpl_imagelist_new() ;

    /* Loop on all planes */
    for (i=0 ; i < nplanes ; i++) {
        image = cpl_image_load_window(filename, CPL_TYPE_T, i, extnum,
                llx, lly, urx, ury);
        if (image == NULL) break;
        
        if (cpl_imagelist_set(imlist, image, i)) break;
        image = NULL;
    }
    if (imlist->ni < nplanes) {
        cpl_image_delete(image);
        cpl_imagelist_delete(imlist);
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }
    
    /* All images in the created imagelist have the same size and type */
    assert( cpl_imagelist_is_uniform(imlist) == 0);
    
    /* Return */
    return imlist ;

#elif CPL_OPERATION == CPL_IMLIST_IO_SAVE

    case CPL_TYPE_T:
        qd.ptype         = CPL_PTYPE;
        qd.CPL_QFITS_BUF = (CPL_TYPE*)CPL_IMAGE_GET(outilist->images[i]) ;
        break ;
    
#endif

#undef CPL_PTYPE
#undef CPL_IMAGE_GET
#undef CPL_TYPE_T
#undef CPL_TYPE
#undef CPL_QFITS_BUF
#undef CPL_IMAGE_WRAP
