/* $Id: cpl_init.c,v 1.10 2005/10/06 12:25:30 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2005 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2005/10/06 12:25:30 $
 * $Revision: 1.10 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <qfits.h>

#include "cpl_memory_impl.h"
#include "cpl_msg.h"
#include "cpl_tools.h"


/**
 * @defgroup cpl_init Library Initialization
 *
 * The module provides the CPL library startup routine. The startup routine
 * initialises CPL internal data structures. For this reason, any application
 * using functions from the CPL libraries @b must call the startup routine
 * prior to calling any other CPL function.
 *
 * @par Synopsis:
 * @code
 *   #include <cpl_init.h>
 * @endcode
 */

/**@{*/

/**
 * @brief
 *   Initialise the CPL core library.
 *
 * @return Nothing.
 *
 * This function sets up the library internal subsystems, which other 
 * CPL functions expect to be in a defined state. In particular, the CPL
 * memory management and the CPL messaging systems are initialised by 
 * this function call.
 *
 * The function must be called once before any other CPL
 * function is used.
 */

void
cpl_init(void)
{

    cpl_memory_init();
    cpl_msg_init();
    return;

}

/**
 * @brief
 *   Stop the internal subsystems of CPL.
 *
 * @return Nothing.
 * @note   Currently, the behaviour of any CPL function becomes
 * undefined after this function is called.
 *
 * This function must be called after any other CPL function
 * is called.
 * 
 */

void
cpl_end(void)
{

    (void) cpl_tools_get_cputime(CPL_CLOCK_RESET);

    cpl_msg_stop();

    qfits_cache_purge();

    return;

}

/**@}*/
